/*
 * ghttp.c
 *
 * Copyright (C) 1993-1997, John Kilburg <john@cs.unlv.edu>
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 */

#include "port_before.h"

#include <stdio.h>

#ifdef HAVE_STRING_H
#include <string.h>
#endif

#ifdef HAVE_STDLIB_H
#include <stdlib.h>
#endif

#include "port_after.h"

struct HTTPContextP
{
  MemPool mp;
  GList fields;         /* list of header fields */

  size_t doff;          /* offset to data */
  bool dfound;          /* data found */

  byte *b;              /* buffer being parsed */
  size_t blen;          /* length of buffer */

  bool foundnl;         /* newline located, for header search */
  size_t nli;           /* newline index, for header search */
  size_t nllen;         /* newline search len */

  size_t di;            /* data index for data callback */

  int status;           /* HTTP status value */
};

static void CheckData _ArgProto((HTTPInfo *));

/*
 * HTTPCheckHeader
 *
 * Scrounges around in the header fields to see if there is any
 * interesting information.
 */
static int
HTTPCheckHeader(hi)
HTTPInfo *hi;
{
  bool cache;
  char *value;
  size_t clen;
  char *option;
  char *list;
  URLParts *up;
  char *url;
  int status;

  status = MIMEGetHTTPStatus(hi->mc);
  if (status >= 300) cache = false;
  else cache = true;

  /*
   * Take action on MIME fields.
   */
  if ((value = MIMEGetField(hi->mc, "location")) != NULL &&
      status >= 300 && status < 400)
  {
    up = URLParse(hi->mp, value);
    
    /* Just blindly follow the URL unless it should be used as a proxy */
    if (status != 305)
    {
      hi->url = MPStrDup(hi->mp, value);
      hi->up = up;
    }
    else hi->pup = up;
    
    return(-1);
  }
  if ((value = MIMEGetField(hi->mc, "content-length")) != NULL)
  {
    clen = (size_t)atoi(value) + hi->doff;
    if (clen > 0 && clen > hi->blen && clen > hi->bsize)
    {
      hi->b = (byte *)realloc_mem(hi->b, clen);
      hi->bsize = clen;
      hi->bmax = clen;
    }
    else hi->bmax = 0;
  }
  if ((value = MIMEGetField(hi->mc, "pragma")) != NULL)
  {
    list = value;
    while ((option = mystrtok(list, ' ', &list)) != NULL)
    {
      if (strcasecmp(option, "no-cache") == 0) cache = false;
    }
  }
  if ((value = MIMEGetField(hi->mc, "www-authenticate")) != NULL)
  {
    cache = false;
  }
  
  if (hi->url != NULL) url = hi->url;
  else if (hi->wr->url != NULL) url = hi->wr->url;
  else url = "unknown";
  
  return(0);
}

/*
 * HTTPParseHeader
 */
static void
HTTPParseHeader(hi)
HTTPContext hi;
{
  size_t doff = hi->doff;
  char *cp, *pcp, *lcp, *fcp;
  bool needstatus;
  bool foundcolon;
  MIMEField *mf;

  if (hi->doff > 0)
  {
    /*
     * Get rid of the HTTP status line.
     */
    needstatus = true;
    for (cp = hi->b; *cp != '\0'; cp++)
    {
      if (*cp == ' ' && needstatus)
      {
	hi->status = atoi(cp);
	needstatus = false;
      }
      else if (*cp == '\n') break;
    }

    /*
     * Start parsing MIME-like fields.
     */
    fcp = NULL;
    pcp = cp;
    lcp = hi->b + hi->doff;
    for (; cp < lcp; cp)
    {
      if (*cp == '\n')
      {
	if (fcp == NULL)
	{
	  mf = MPGet(hi->mp, sizeof(MIMEField));
	  mf->name = MPGet(hi->mp, cp - pcp + 1);
	  strncpy(mf->name, pcp, cp - pcp);
	  mf->name[cp - pcp] = '\0';
	  mf->value = "";
	}
	else
	{
	  mf = MPGet(hi->mp, sizeof(MIMEField));
	  mf->name = MPGet(hi->mp, fcp - pcp + 1);
	  strncpy(mf->name, pcp, fcp - pcp);
	  mf->name[fcp - pcp] = '\0';

	  for (fcp++; fcp < cp && isspace8(*fcp); fcp++)
	      ;

	  if (fcp == cp) mf->value = "";
	  else
	  {
	    mf->value = MPGet(hi->mp, cp - fcp + 1);
	    strncpy(mf->value, fcp, cp - fcp);
	    mf->value[cp - fcp] = '\0';
	  }
	  fcp = NULL;
	}
	GListAddTail(hi->fields, mf);
	pcp = cp + 1;
      }
      else if (cp[i] == ':' && fcp == NULL) fcp = cp;
    }
  }

  return;
}

/*
 * HTTPHandleHeader
 */
static void
HTTPHandlerHeader(hi)
HTTPContext hi;
{
  size_t i;
  byte *b = hi->b;
  size_t blen = hi->blen;

  /*
   * Check for non-HTTP/1.x.
   */
  if ((blen >= 1 && b[0] != 'H') || (blen >= 2 && b[1] != 'T') ||
      (blen >= 3 && b[2] != 'T') || (blen >= 4 && b[3] != 'P'))
  {
    hi->dfound = true;
    hi->doff = 0;
    HTTPHandleData(hi);
    return;
  }
  
  /*
   * Check for end of header/beginning of data.
   */
  for (i = hi->nllen; i < blen; i++)
  {
    if (b[i] == '\n')
    {
      if (hi->nlfound)
      {
	if (hi->nli == i - 1 || (hi->nli == i - 2 && b[i - 1] == '\r'))
	{
	  hi->doff = i + 1;
	  hi->dfound = true;
	  HTTPParseHeader(hi);
	  break;
	}
	else hi->nli = i;
      }
      else
      {
	hi->nlfound = true;
	hi->nli = i;
      }
    }
  }
  hi->nllen = blen;
  
  return;
}

/*
 * HTTPHandleData
 */
static void
HTTPHandleData(hi)
HTTPContext hi;
{
  return;
}

/*
 * HTTPAddData
 */
void
HTTPAddData(hi, b, blen)
HTTPContext hi;
byte *b;
size_t blen;
{
  myassert(hi->blen <= blen, "hi->blen > blen!");

  if (blen == 0) return;

  hi->b = b;
  hi->blen = blen;

  if (hi->dfound) HTTPHandleData(hi);
  else HTTPHandleHeader(hi);

  return;
}

/*
 * HTTPEndData
 */
void
HTTPEndData(hi, b, blen)
HTTPContext hi;
byte *b;
size_t blen;
{
  myassert(hi->blen <= blen, "hi->blen > blen!");

  if (blen == 0) return;

  hi->b = b;
  hi->blen = blen;

  if (hi->dfound) HTTPHandleData(hi);
  else HTTPHandleHeader(hi);

  return;
}

/*
 * HTTPDestroyContext
 */
void
HTTPDestroyContext(hi)
HTTPContext hi;
{
  MPDestroy(hi->mp);
  return;
}

/*
 * HTTPCreateContext
 */
HTTPContext
HTTPCreateContext(dataproc, headerproc, errorproc, closure)
HTTPDataProc dataproc;
HTTPHeaderProc headerproc;
HTTPErrorProc errorproc;
void *closure;
{
  HTTPContext hi;
  MemPool mp;

  mp = MPCreate();
  hi = (HTTPContext)MPCGet(mp, sizeof(struct HTTPContextP));
  hi->mp = mp;
  hi->fields = GListCreateX(mp);

  return(hi);
}
